#!/bin/bash

#DESCRIPTION: write boot and .dtb file to boot flash at 7Mb (or 5Mb-384kb) seek

# Copyright 2017-2024 AO MCST
# Distributed under the terms of MIT License.

VER="0.9.2"

if [[ ! -z ${VERBOSE} ]]; then
	VERBOSE=${VERBOSE}
fi

BAK=0
ZERO=0
WRITE=1
CHKKOMBI=1
spidevflag=0

# board specific
FL_DEVICE="-p linux_spi:dev=/dev/spidev0.0,spispeed=12500"

SEEK=7340032
case `uname -p` in
	E2C3)
		SEEK=4849664
		;;
	E16C)
		SEEK=4849664
		;;
esac


check_file() {
	if [[ ! -f "$1" ]]; then
		echo "File $1 not found"
		exit 1
	fi
}

check_ret() {
	if [[ $1 -ne 0 ]]; then
		echo "$2 error"
		return_spidev
		exit 1
	fi
}

return_spidev() {
	if [ "$spidevflag" -eq 1 ]; then
		echo "spi0.0" > /sys/bus/spi/drivers/spidev/unbind || { echo "Error: Failed to unbind driver spidev.";}
		echo spi-nor > /sys/bus/spi/devices/spi0.0/driver_override || { echo "Error: Failed to override driver.";}
		echo "spi0.0" > /sys/bus/spi/drivers/spi-nor/bind || { echo "Error: Failed to bind driver spi-nor."; }
	fi
}

chk_spidev() {
	lsmod | grep -q spidev
	check_ret $? "Find spidev"
	lines=$(cat /proc/mtd | wc -l)
	if [ "$lines" -ge 2 ]; then
		cat /proc/mtd | tail -n +2
		echo "MTD usage detected, use spidev and overwrite flash?"
		read -p "yes/no: " confirm
		if [ "$confirm" == "yes" ]; then
			echo "spi0.0" > /sys/bus/spi/drivers/spi-nor/unbind || { echo "Error: Failed to unbind driver spi-nor.";}
			echo spidev > /sys/bus/spi/devices/spi0.0/driver_override || { echo "Error: Failed to override driver.";}
			echo "spi0.0" > /sys/bus/spi/drivers/spidev/bind || { echo "Error: Failed to bind driver spi-nor."; }
			spidevflag=1
		else
			echo "spidev will not be used"
		fi
	fi
}

generate_zero_dtb() {
	echo "Clear boot flash area from offset $2"
	(($VERBOSE)) && echo "dd if=/dev/zero of=${zerofile} count=100 bs=1K | tr \"\\000\" \"\\377\" > ${zerofile}"
	dd if=/dev/zero count=100 bs=1K | tr "\000" "\377" > ${zerofile} 2>/dev/null
	check_ret $? "dd generate zero-file"
	(($VERBOSE)) && echo "dd bs=1c seek=$2 if=${zerofile} of=$1 conv=notrunc"
	dd bs=1c seek=$2 if=${zerofile} of=$1 conv=notrunc >/dev/null 2>&1
	check_ret $? "dd zero write"
}

generate_dump() {
	s=$(stat -c "%s" $1)
	echo "Put $1 (size ${s}) to flash area from offset $3"
	(($VERBOSE)) && echo "dd bs=1c seek=$3 if=$1 of=$2 conv=notrunc"
	dd bs=1c seek=$3 if=$1 of=$2 conv=notrunc >/dev/null 2>&1
	check_ret $? "dd dump write"
}

write_dump() {
	echo "Write image to flash chip"
	(($VERBOSE)) && echo "${FL} -c \"${flash_chip}\" ${FL_DEVICE} -w $1"
	${FL} -c "${flash_chip}" ${FL_DEVICE} -w $1 >/dev/null
	check_ret $? "Flashrom write"
}

load_module() {
	loadcount=0
	BFS=`cat /sys/module/spidev/parameters/bufsiz 2>/dev/null`
	if [[ x"${BFS}" != x'64' ]]; then
		echo "spidev bufsiz change to 64 (old: ${BFS})"
		modprobe -r spidev
		while [ -d /sys/module/spidev ]; do
			sleep 1
			(( loadcount = $loadcount + 1 ))
			if [[ $loadcount -gt 10 ]]; then
				break
			fi
		done
		modprobe spidev bufsiz=64
		check_ret $? "Init kernel module"
	fi
}


echo "Script version: ${VER}"

command -v flashrom 2>&1 >/dev/null
check_ret $? "Find flashrom"
FL=`command -v flashrom`

# print help if no params
if [[ -z $1 || $1 == "-h" ]]; then
	echo "Write boot (.bin) and/or devicetree (.dtb) images to boot flash."
	echo "Usage:"
	echo "  `basename $0` <boot.bin>            - backup and write boot to flash"
	echo "  `basename $0` <file.dtb>            - backup and write devicetree to flash"
	echo "  `basename $0` <boot.bin> <file.dtb> - backup and write boot+devicetree to flash"
	echo "  `basename $0` -i                    - only show info about flash"
	echo "  `basename $0` -b                    - only backup flash"
	echo "  `basename $0` -z                    - backup and clear devicetree in flash"
	echo "  `basename $0` -z <file.dtb>         - backup, clear and write devicetree to flash"
	echo "  `basename $0` -k <image.bin>        - backup and write full or Kombinator image to flash"
	exit 0
fi

load_module

# parse params
bootfile=
dtbfile=

opt=$1
shift

if [[ ${opt} == "-i" ]]; then
	(($VERBOSE)) && echo "${FL} ${FL_DEVICE} --flash-name"
	echo "Probe for boot flash chip..."
	${FL} ${FL_DEVICE} 2>/dev/null | grep "Found"
	exit 0
fi

case ${opt} in
-z )
	ZERO=1
	if [[ -z $1 ]]; then
		WRITE=0
	else
		dtbfile=$1
		check_file ${dtbfile}
	fi
	;;
-b )
	BAK=1
	;;
-k )
	CHKKOMBI=0
	bootfile=$1
	check_file ${bootfile}
	;;
*  )
	bootfile=${opt}
	check_file ${bootfile}

	if [[ ! -z ${bootfile} ]] && [[ ! -z $1 ]]; then
		dtbfile=$1
		check_file ${dtbfile}
	else
		echo ${bootfile} | grep "\.dtb$" >/dev/null
		if [[ $? -eq 0 ]]; then
			dtbfile=${bootfile}
			bootfile=
		fi
	fi
	;;
esac

## flashrom (spidev) or MTD (spi-nor)
chk_spidev

## probe for flash chip
flash_chip=""
(($VERBOSE)) && echo "${FL} ${FL_DEVICE} | grep \"Found \" | grep \" | cut -d\" -f2"
chips=$(${FL} ${FL_DEVICE} 2>/dev/null | grep "Found " | grep \" | cut -d'"' -f2)

nchips=$(echo "${chips}" | grep -c '^')
if [[ -z ${chips} ]]; then
	echo "No flash chip found"
	return_spidev
	exit 1
elif [[ ${nchips} -eq 1 ]]; then
	flash_chip=${chips}
else
	echo "Found flash chips"
	OLD_IFS=$IFS
	IFS=$'\n'
	PS3='Select flash chip (enter number): '
	select chip in ${chips[@]}
	do
		case ${chip} in
		*)
			flash_chip=${chip}
			break
			;;
		esac
	done
	IFS=$OLD_IFS
fi

if [[ -z "${flash_chip}" ]]; then
	echo "Unsupported flash chip ${flash_chip}"
	return_spidev
	exit 1
fi
echo "Use flash chip: ${flash_chip}"

# read flash
bakfile=`pwd`"/"`date +%Y-%m-%d_%H:%M:%S`_flashimage.bak
echo "Save current flash to ${bakfile}"
(($VERBOSE)) && echo "${FL} -c \"${flash_chip}\" ${FL_DEVICE} -r \"${bakfile}\""
${FL} -c "${flash_chip}" ${FL_DEVICE} -r "${bakfile}" >/dev/null

# check for Kombinator
kombsign=`hexdump -s 0xFFF000 -n 4 -e '"%02X"' ${bakfile} 2>/dev/null`
if [[ x${kombsign} == x'B007E2E2' ]]; then
	echo "Found Kombinator signature: ${kombsign}"
else
	echo "Kombinator signature not found (${kombsign})"
fi

# only backup flash
if [[ ${BAK} == "1" ]]; then
	return_spidev
	echo "Done"
	exit 0
fi

# print help and exit if Kombinator found
if [[ ${CHKKOMBI} == "1" ]]; then
	if [[ x${kombsign} == x'B007E2E2' ]]; then
		echo "Use Kombinator to generate flash dump."
		echo "For flash write use command:"
		echo "${FL} -c "${flash_chip}" ${FL_DEVICE} -w <dumpfile>"
		exit 1
	fi
fi

# temporary file
tmpfile=`tempfile 2>/dev/null` || tmpfile=/tmp/spidev00flashrom$$
zerofile=`tempfile 2>/dev/null` || zerofile=/tmp/spidev00zero$$
# make it autoremove
trap "rm -f ${tmpfile} ${zerofile}" 0 1 2 5 15

# change filesize and create file for write
cp "${bakfile}" ${tmpfile}

## generate and write

# clear devicetree in flash
if [[ ${ZERO} == "1" ]]; then
	generate_zero_dtb ${tmpfile} ${SEEK}
	write_dump ${tmpfile}
fi

# write boot and/or devicetree to flash
if [[ ${WRITE} == "1" ]]; then
	if [[ ! -z ${bootfile} ]]; then
		generate_dump ${bootfile} ${tmpfile} 0
	fi
	if [[ ! -z ${dtbfile} ]]; then
		generate_dump ${dtbfile} ${tmpfile} ${SEEK}
	fi
	write_dump ${tmpfile}
fi

return_spidev
echo "Done"
exit 0
